from ribosome.machine import may_handle
from ribosome.unite import UniteMessage, UniteSource, UniteKind

from amino import Map, List

from myo.plugins.unite.message import UniteHistory
from myo.state import MyoTransitions, MyoComponent


class UniteTransitions(MyoTransitions):

    def unite_cmd(self, cmd):
        args = ' '.join(self.msg.unite_args)
        self.vim.cmd('Unite {} {}'.format(cmd, args))

    @may_handle(UniteHistory)
    def history(self):
        self.unite_cmd(UniteNames.history)


class Plugin(MyoComponent):
    Transitions = UniteTransitions

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._unite_ready = False

    def prepare(self, msg):
        if not self._unite_ready and isinstance(msg, UniteMessage):
            self._setup_unite()

    def _setup_unite(self):
        history = UniteSource(UniteNames.history,
                              UniteNames.history_candidates,
                              UniteNames.history)
        run_action = Map(name='run', handler=UniteNames.run,
                         desc='run command')
        run = UniteKind(UniteNames.history, List(run_action))
        history.define(self.vim)
        run.define(self.vim)
        self._unite_ready = True


class UniteNames():
    history_candidates = '_myo_unite_history'

    run = '_myo_unite_run_project'

    history = 'myo_history'

__all__ = ('Plugin', 'UniteNames')



from typing import Callable, Any

from amino.task import Task

from lenses import Lens, lens
from amino import Either, List, F, _, L, Left, __, Right, I, Just, Boolean
from amino.lens.tree import path_lens_unbound

from ribosome.record import field, list_field, map_field
from ribosome.machine import Message

from myo.ui.tmux.window import Window
from myo.logging import Logging
from myo.ui.tmux.view import View
from myo.ui.tmux.layout import Layout
from myo.ui.tmux.pane import Pane
from myo.ui.tmux.session import Session
from myo.ui.tmux.data import TmuxState
from myo.record import Record


class ViewPath(Record):
    session = field(Session)
    window = field(Window)
    view = field(View)
    layout = field(Layout)
    outer = list_field()

    @staticmethod
    def create(session, window, view, layout, outer):
        return ViewPath(session=session, window=window, view=view,
                        layout=layout, outer=outer)

    @staticmethod
    def try_create(session: Session, window: Window, views: List[View]
                   ) -> Either[str, 'ViewPath']:
        def try_create1(view, layouts):
            return (
                layouts.detach_last
                .to_either('ViewPath.create: only two items in views')
                .map2(L(ViewPath.create)(session, window, view, _, _))
            )
        return (
            views.detach_last.to_either('ViewPath.create: empty List[View]')
            .flat_map2(try_create1)
        )

    @property
    def layouts(self):
        return self.outer.cat(self.layout)

    @property
    def view_list(self):
        return self.layouts.cat(self.view)

    def map(self, fp: Callable[[View], View], fl: Callable[[Layout], Layout]):
        return ViewPath.create(
            self.session, self.window, fp(self.view), fl(self.layout),
            self.outer / fl
        )

    def map_view(self, f: Callable[[View], View]):
        return self.map(f, I)

    @property
    def is_pane(self):
        return Boolean(isinstance(self.view, Pane))

    @property
    def loc(self):
        return ViewLoc.create(self.session, self.window, self.view)

PPTrans = Callable[[ViewPath], Task[Either[Any, ViewPath]]]


class ViewPathLens(Record):
    lens = field(Lens)

    @staticmethod
    def create(lens: Lens):
        return ViewPathLens(lens=lens)

    def _check_result_type(self, result):
        if (
                not isinstance(result, Either) or (
                    result.is_right and
                    not result.exists(L(isinstance)(_, ViewPath))
                )
        ):
            msg = ('invalid ViewPath transition result type, must be' +
                   ' Either[..., ViewPath]: {}')
            raise Exception(msg.format(result))

    def run(self, state: TmuxState, f: PPTrans) -> Task[Either[Any, Window]]:
        bound = self.lens.bind(state)
        session, (window, views) = bound.get()
        path = ViewPath.try_create(session, window, List.wrap(views))
        set = lambda a: bound.set((session, (window, a)))
        def result(r):
            self._check_result_type(r)
            return r / _.view_list / set
        return Task.from_either(path) // F(f) / result


def _initial_ppm_f(pp, window) -> Task[Either[Any, Window]]:
    return Task.now(Right(window))


class ViewPathMod(Logging, Message):
    pred = field(Callable)
    _f = field(Callable, initial=(lambda: _initial_ppm_f))
    options = map_field()

    @property
    def attr(self):
        return (lambda a: self._pane_attr(a).or_else(self._layout_attr(a)))

    @property
    def _pane_attr(self):
        return self._attr(_.panes)

    @property
    def _layout_attr(self):
        return self._attr(_.layouts)

    def _attr(self, f):
        return (
            f(__)
            .find_lens_pred(self.pred)
            .map(f(lens()).add_lens)
        )

    def _chain(self, f: PPTrans, g: Callable):
        h = lambda vp: L(vp.run)(_, f)
        chain = lambda vp, win: self._f(vp, win) // g(win, h(vp))
        return type(self)(pred=self.pred, _f=chain)  # type: ignore

    def map(self, f: PPTrans) -> Task[Either[Any, Window]]:
        keep_error = F(Left) >> Task.now
        return self._chain(f, lambda w, d: __.cata(keep_error, d))

    __truediv__ = map

    def and_then(self, f: PPTrans):
        return self._chain(f, lambda w, d: lambda a: d(a | w))

    __add__ = and_then

    def foreach(self, f: PPTrans):
        run = lambda path: f(path).replace(Right(path))
        return self.map(run)

    __mod__ = foreach

    def run(self, state: TmuxState) -> Task:
        return self._lens(state) // L(self._f)(_, state)

    def _lens(self, state):
        err = 'view lens path failed for {}'.format(self.pred)
        def main_lens(w):
            return ((Just(lens()) &
                     path_lens_unbound(w.root, _.layouts, self.attr)
                     .map(lens().root.add_lens)
                     )
                    .map2(lens().tuple_))
        def s_lens(s):
            return ((Just(lens()) & s.attr_lens(_.windows, main_lens))
                    .map2(lens().tuple_))
        st_lens = state.attr_lens(_.sessions, s_lens)
        return Task.from_maybe(st_lens, err) / ViewPathLens.create


class PanePathMod(ViewPathMod):

    @property
    def attr(self):
        return self._pane_attr


class LayoutPathMod(ViewPathMod):

    @property
    def attr(self):
        return self._layout_attr


def ppm_id(path: ViewPathMod):
    return Task.now(Right(path))


class ViewLoc(Record):
    view = field(View)
    window = field(Window)
    session = field(Session)

    @staticmethod
    def create(session, window, view):
        return ViewLoc(view=view, window=window, session=session)

    @property
    def _str_extra(self):
        return List(self.session, self.window, self.view)

    def with_view(self, v: View):
        return self.set(view=v)


class LayoutLoc(ViewLoc):
    view = field(Layout)

__all__ = ('ViewPath', 'ViewPathLens', 'ViewPathMod', 'ppm_id')



from functools import singledispatch  # type: ignore

from myo.command import Command, TransientCommand


@singledispatch
def unite_format(obj):
    return str(obj)


@unite_format.register(Command)
def unite_format_command(cmd):
    return dict(word=cmd.name, name=cmd.name)


@unite_format.register(TransientCommand)
def unite_format_transient_command(cmd):
    line = '{} {}'.format(cmd.command.name, cmd.line)
    return dict(word=line, name=cmd.name)

__all__ = ('unite_format',)



from numbers import Number

from ribosome.record import maybe_field, field, bool_field

from amino import Just, L, List, _

from myo.record import Named


class View(Named):
    name = field(str)
    min_size = maybe_field(Number)
    max_size = maybe_field(Number)
    fixed_size = maybe_field(Number)
    minimized = bool_field()
    minimized_size = maybe_field(Number)
    weight = maybe_field(Number)
    position = maybe_field(Number)

    @property
    def _minimized_size(self):
        return self.minimized_size | 2

    @property
    def effective_weight(self):
        return (Just(0) if self.fixed_size.is_just or self.minimized else
                self.weight)

    @property
    def effective_fixed_size(self):
        return self.minimized.cata(Just(self._minimized_size), self.fixed_size)

    @property
    def size_desc(self):
        candidates = List(
            ('min', self.min_size), ('max', self.max_size),
            ('fix', self.fixed_size), ('weight', self.weight),
            ('pos', self.position), ('mini', self.minimized_size)
        )
        def fmt(a, b):
            return b / L('| {} -> {}'.format)(a, _)
        return candidates.flat_map2(fmt).mk_string(' ')

__all__ = ('View',)



from typing import Callable, Tuple

from amino import Maybe, __, _, Just, Task, Either, F, L, Right, List, Map

from myo.logging import Logging
from myo.ui.tmux.data import TmuxState
from myo.ui.tmux.view import View
from myo.ui.tmux.pane import Pane
from myo.ui.tmux.layout import Layout
from myo.util import Ident
from myo.ui.tmux.server import Server, NativeServer
from myo.ui.tmux.view_path import (ViewLoc, ViewPath, PanePathMod,
                                   LayoutPathMod, ViewPathMod)
from myo.ui.tmux.session import Session
from myo.ui.tmux.window import Window
from myo.ui.tmux.facade.window import WindowFacade


class TmuxFacade(Logging):

    def __init__(self, state: TmuxState, socket=None, options: Map=Map()
                 ) -> None:
        self.state = state
        self.socket = socket
        self.options = options

    @property
    def native_server(self):
        return NativeServer(socket_name=self.socket)

    @property
    def _fresh_server(self):
        return Server(self.native_server)

    @property
    def server(self):
        return self._fresh_server

    @property
    def sessions(self):
        return self.server.sessions

    def native_session(self, id: int):
        return self.sessions.find(__.id_i.contains(id))

    def native_window(self, session, window):
        win = lambda s: window.id // s.window_by_id
        return ((session.id // self.native_session // win)
                .to_either('no native window for {}'.format(window)))

    def _view_loc(self, ident: Ident, attr: Callable, desc: str, tpe: type
                  ) -> Either[str, ViewLoc]:
        win = lambda w: Just(w) & attr(w.root)(ident)
        sess = lambda s: s.windows.find_map(win).map2(lambda a, b: (s, a, b))
        return (
            self.state.sessions
            .find_map(sess)
            .to_either('{} not found: {}'.format(desc, ident))
            .map3(ViewLoc.create)
        )

    def pane_loc(self, ident: Ident) -> Either[str, ViewLoc]:
        return self._view_loc(ident, _.find_pane, 'pane', ViewLoc)

    def view_loc(self, ident: Ident) -> Either[str, ViewLoc]:
        return self._view_loc(ident, _.find_view, 'view', ViewLoc)

    def window(self, session: Session, window: Window):
        return (self.native_window(session, window) /
                L(WindowFacade)(session, window, _))

    def window_task(self, session, window) -> Task[WindowFacade]:
        return Task.call(self.window, session, window).join_either

    def loc_window(self, loc: ViewLoc) -> WindowFacade:
        return self.window(loc.session, loc.window)

    def add_to_layout(self, parent: Maybe[str], target: Callable, view: View):
        name = parent | 'root'
        return (
            self.state.layout_lens_ident(name) /
            target /
            __.modify(__.cat(view))
        )

    def add_pane_to_layout(self, parent: Maybe[str], pane: Pane):
        return self.add_to_layout(parent, _.panes, pane)

    def add_layout_to_layout(self, parent: Maybe[str], layout: Layout):
        return self.add_to_layout(parent, _.layouts, layout)

    def pane_window(self, ident: Ident
                    ) -> Either[str, Tuple[WindowFacade, ViewLoc]]:
        f = lambda loc: self.loc_window(loc) & Right(loc.view)
        return self.pane_loc(ident) // f

    def pane_window_task(self, ident: Ident
                         ) -> Task[Tuple[WindowFacade, ViewLoc]]:
        return Task.call(self.pane_window, ident).join_either

    def view_window(self, ident: Ident
                    ) -> Either[str, Tuple[WindowFacade, ViewLoc]]:
        f = lambda loc: self.loc_window(loc) & Right(loc.view)
        return self.view_loc(ident) // f

    def path_window(self, path: ViewPath) -> Maybe[WindowFacade]:
        return self.window(path.session, path.window)

    def path_window_task(self, path: ViewPath) -> Task[WindowFacade]:
        return self.window_task(path.session, path.window)

    @property
    def native_vim_window(self):
        return ((self.state.vim_session & self.state.vim_window)
                .flat_map2(self.native_window))

    def focus(self, ident: Ident):
        return self._focus(ident, Just)

    def fix_focus(self, ident: Ident, alt: Ident):
        check = lambda a: a.view.focus.c(Just(a), F(self.pane_loc, alt))
        return self._focus(ident, check)

    def _focus(self, ident: Ident, f: Callable):
        run = lambda l: self.loc_window(l) / __.focus(l.view)
        return self.pane_loc(ident) // f // run

    @property
    def pack_sessions(self):
        return (self.state.sessions / self.pack_windows).sequence(Task)

    def pack_windows(self, session: Session):
        return session.windows.traverse(L(self.pack_window)(session, _), Task)

    def pack_window(self, session: Session, window: Window
                    ) -> Task[Either[str, Window]]:
        return self.window_task(session, window) // __.pack()

    def open_pane(self, path: ViewPath) -> Task[Either[str, ViewPath]]:
        return self.path_window_task(path) // __.open_pane(path)

    def kill_process(self, ident: Ident, signals):
        return (
            self.pane_window_task(ident)
            .flat_map2(__.kill_process(_, signals))
        )

    @property
    def close_all(self):
        return (
            self.state.possibly_open_panes //
            _.id
        ).traverse(self.close_id, Task)

    def close_id(self, id: int):
        return self.server.cmd('kill-pane', '-t', '%{}'.format(id))

    def view_open(self, ident: Ident):
        return self.view_window(ident).map2(lambda w, v: w.view_open(v))

    def _run_vpm(self, vpm):
        return self._state_task(vpm.run, self.options)

    def _ident_pm(self, tpe: type, ident: Ident):
        return tpe(pred=__.has_ident(ident), options=self.options)

    def _ident_ppm(self, ident: Ident):
        return self._ident_pm(PanePathMod, ident)

    def _ident_lpm(self, ident: Ident):
        return self._ident_pm(LayoutPathMod, ident)

    def _ident_vpm(self, ident: Ident):
        return self._ident_pm(ViewPathMod, ident)

    def open_pane_ppm(self, name: Ident):
        return self._ident_vpm(name) / self.open_pane

    def close_pane_ppm(self, ident: Ident):
        def run(w, v):
            return self._ident_ppm(ident) / w.close_pane_path
        return self.pane_window(ident).map2(run)

    def run_command_ppm(self, pane_ident: Ident, line: str, in_shell: bool,
                        kill: bool, signals: List[str]):
        win = lambda path: self.path_window_task(path)
        def check_running(path):
            pane_kill = path.view.kill
            return Task.now(Right(path)) if in_shell else (
                win(path) //
                __.ensure_not_running(
                    path.view, kill=kill or pane_kill, signals=signals) /
                __.replace(path)
            )
        def pipe(path):
            return (
                win(path) //
                __.pipe(path.view, self.state.instance_id) /
                path.setter.view /
                Right
            )
        run = lambda path: (win(path) // __.run_command_line(path.view, line))
        def pid(path):
            return (
                win(path) //
                __.command_pid(path.view) /
                (lambda a: path.map_view(__.set(pid=a.to_maybe))) /
                Right
            )
        ppm = (self._ident_vpm
               if self.pane_open(pane_ident) else
               self.open_pane_ppm)
        return (((ppm(pane_ident) + check_running) / pipe) % run) / pid

    def pane_mod(self, ident: Ident, f: Callable[[Pane], Pane]):
        return self._mod(self._ident_ppm, ident, f)

    def view_mod(self, ident: Ident, f: Callable[[View], View]):
        return self._mod(self._ident_vpm, ident, f)

    def _mod(self, ctor: Callable, ident: Ident, f: Callable[[View], View]):
        def g(path):
            return Task.now(Right(path.map_view(f)))
        return ctor(ident) / g

    def pinned_panes(self, ident):
        def layout(l):
            target = l.panes.find(__.has_ident(ident))
            sub = (l.layouts // layout)
            add = (List() if sub.empty and target.empty else
                   l.panes.filter(_.pin))
            return sub + add
        return (self.state.all_windows / _.root // layout) / _.ident

    def pane_open(self, ident: Ident):
        return self.pane_window(ident).map2(__.pane_open(_)).true

    def view_exists(self, ident: Ident):
        return self.view_loc(ident).present

__all__ = ('TmuxFacade',)
